from __future__ import annotations

import warnings
from io import BytesIO
from logging import getLogger
from pathlib import Path
from typing import Literal

import aiofiles
import joblib
from pandas import DataFrame, concat
from pandas.testing import assert_frame_equal

BASE_PATH = Path(f"~/.cache/{__name__.split('.')[0]}").expanduser()
LOG = getLogger(__name__)


def get_path(folder: str, name: str) -> Path:
    """Get path to cache file.

    Parameters
    ----------
    folder : str
        The folder name.
    name : str
        The file name.

    Returns
    -------
    Path
        The path to cache file.
    """
    return BASE_PATH / folder / f"{name}.lz4"


async def read(path: Path) -> DataFrame:
    """Read cache file using joblib and aiofiles, and return DataFrame.
    If cache file does not exist, return empty DataFrame.

    Parameters
    ----------
    path : Path
        The path to cache file.

    Returns
    -------
    DataFrame
        The DataFrame read from cache file.
    """
    if path.exists():
        async with aiofiles.open(path, "rb") as f:
            with BytesIO(await f.read()) as f2:
                return joblib.load(f2)
    return DataFrame()


async def save(
    path: Path,
    df: DataFrame,
    *,
    compress: int | str | tuple[str, int] = ("lz4", 3),
    protocol: int | None = None,
) -> None:
    """Save DataFrame to cache file using joblib and aiofiles.

    Parameters
    ----------
    path : Path
        The path to cache file.
    df : DataFrame
        The DataFrame to save.
    compress : int | str | tuple[str, int], optional
        The compression level, by default ("lz4", 3)
    protocol : int | None, optional
        The pickle protocol, by default None (latest protocol)
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    with BytesIO() as f:
        joblib.dump(df, f, compress=compress, protocol=protocol)
        async with aiofiles.open(path, "wb") as f2:
            await f2.write(f.getvalue())


async def update(
    path: Path,
    df: DataFrame,
    *,
    df_old: DataFrame | None = None,
    reload: bool = False,
    mismatch: Literal["warn", "raise"] | int | None = "warn",
    keep: Literal["first", "last"] = "last",
    compress: int | str | tuple[str, int] = ("lz4", 3),
    protocol: int | None = None,
) -> DataFrame:
    """Update cache file with DataFrame.

    Parameters
    ----------
    path : Path
        The path to cache file.
    df : DataFrame
        The DataFrame to save.
    df_old : DataFrame, optional
        The DataFrame read from cache file, by default None
        If not None, reload is ignored.
        Intended to avoid reading cache file multiple times.
    reload : bool, optional
        Whether to ignore cache file and reload, by default False
        If df_old is not None, reload is ignored.
    mismatch : Literal["warn", "raise"] | int | None, optional
        The action when data mismatch, by default "warn"
        If int, log level. If None, do nothing.
    keep : Literal["first", "last"], optional
        Which duplicated index to keep, by default "last" (has no effect if mismatch is "raise")
    compress : int | str | tuple[str, int], optional
        The compression level, by default ("lz4", 3)
    protocol : int | None, optional
        The pickle protocol, by default None (latest protocol)

    Returns
    -------
    DataFrame
        The updated DataFrame.
    """
    if df_old is None:
        df_old = await read(path) if not reload else DataFrame()

    # check if duplicated data is same
    if len(df_old) > 0:
        idx = df.index.intersection(df_old.index)
        try:
            assert_frame_equal(df_old.loc[idx], df.loc[idx], rtol=1e-3)
        except AssertionError as e:
            if mismatch == "raise":
                raise e
            elif mismatch == "warn":
                warnings.warn(
                    f"Data mismatch. Left: cache, Right: new\n{e}", RuntimeWarning
                )
            elif isinstance(mismatch, int):
                LOG.log(mismatch, f"Data mismatch. Left: cache, Right: new\n{e}")
            else:
                pass

    # concat
    df = concat([df_old, df], axis=0)
    # drop duplicated index
    df = df[~df.index.duplicated(keep=keep)]
    await save(path, df, compress=compress, protocol=protocol)
    return df



from .base import HistoricalDataCache
from .chunk import HistoricalDataCacheWithChunk, HistoricalDataCacheWithFixedChunk
from .id import IdCacheWithFixedChunk

__all__ = [
    "HistoricalDataCache",
    "HistoricalDataCacheWithChunk",
    "HistoricalDataCacheWithFixedChunk",
    "IdCacheWithFixedChunk",
]



from __future__ import annotations

from abc import ABCMeta
from typing import Any, Generic, Sequence

from pandas import Index

from .base import PGet, TIndex
from .chunk import HistoricalDataCacheWithFixedChunk


class IdCacheWithFixedChunk(
    HistoricalDataCacheWithFixedChunk[TIndex, Any, PGet],
    Generic[TIndex, PGet],
    metaclass=ABCMeta,
):
    """Base class for data cache indexed by ids.

    Examples
    --------
    .. code-block:: python

       from cached_historical_data_fetcher import IdCacheWithFixedChunk

       class MyIdCache(IdCacheWithFixedChunk[str, Any]):
           delay_seconds = 0.0

           async def get_one(self, start: str, *args: Any, **kwargs: Any) -> DataFrame:
               return DataFrame({"id+hello": [start + "+hello"]}, index=[start])

       cache = MyIdCache()
       cache.set_ids(["a", "b", "c"])
       df = await cache.update()

    Output:

       id+hello
    a   a+hello
    b   b+hello
    c   c+hello
    """

    add_interval_to_start_index: bool = False
    subtract_interval_from_end_index: bool = False
    ids: Index[TIndex] | None = None

    @property
    def start_index(self) -> TIndex | None:  # type: ignore
        return None

    @property
    def end_index(self) -> TIndex | None:  # type: ignore
        return None

    def set_ids(self, ids: Sequence[TIndex], *, refresh: bool = False) -> None:
        """Set ids to be fetched.

        Parameters
        ----------
        ids : Sequence[TIndex]
            New ids to be stored
        refresh : bool, optional
            If True, ids will be replaced by the given ids.
            If False, ids will be added to the existing ids, by default False
        """
        if self.ids is None or refresh:
            self.ids = Index(ids)
        self.ids = self.ids.union(Index(ids))

    def new_indices(
        self, start: Any, end: Any, old_indices: Sequence[TIndex]
    ) -> Sequence[TIndex]:
        if self.ids is None:
            raise RuntimeError("ids is not set, call `set_ids()` first")
        return self.ids.difference(Index(old_indices)).to_list()

    def to_update(self, end: TIndex, *args: Any, **kwargs: Any) -> bool:
        return True



from __future__ import annotations

import asyncio
import warnings
from abc import ABCMeta, abstractmethod
from logging import getLogger
from typing import Generic, Sequence

import numpy as np
import pandas as pd
from pandas import DataFrame, Timestamp, concat
from tqdm.auto import tqdm

from .base import HistoricalDataCache, PGet, TIndex, TInterval

LOG = getLogger(__name__)


class HistoricalDataCacheWithChunk(
    HistoricalDataCache[TIndex, TInterval, PGet],
    Generic[TIndex, TInterval, PGet],
    metaclass=ABCMeta,
):
    """Base class for historical data cache with chunk.

    Usage
    -----
    1. Override `self.get_one()` to implement the logic.
    2. Override `self.to_update()` if the index is not Timestamp or interval is not fixed.
    3. Call `self.update()` to get historical data.

    Examples
    --------
    .. code-block:: python
        from cached_historical_data_fetcher import HistoricalDataCacheWithChunk
        from pandas import DataFrame, Timedelta, Timestamp

        class MyCacheWithChunk(HistoricalDataCacheWithChunk[Timestamp, Timedelta, Any]):
            delay_seconds = 0.0
            interval = Timedelta(days=1)
            start_index = Timestamp.utcnow().floor("10D")

            async def get_one(self, start: Timestamp, *args: Any, **kwargs: Any) -> DataFrame:
                return DataFrame({"day": [start.day]}, index=[start])

        df = await MyCacheWithChunk().update()
    """

    delay_seconds: float
    """Delay between chunks in seconds."""

    @property
    def start_index(self) -> TIndex:
        """The start index of historical data.
        Used when cache file exists."""
        raise NotImplementedError()

    def new_indices(
        self, start: TIndex, end: TIndex, old_indices: Sequence[TIndex]
    ) -> Sequence[TIndex]:
        """Get new indices to update.

        Parameters
        ----------
        start : TIndex
            The start index of historical data.
        end : TIndex
            The end index of historical data.
        old_indices : Sequence[TIndex]
            The old indices of historical data

        Returns
        -------
        Sequence[TIndex]
            The new indices to update.

        Raises
        ------
        TypeError
            If `start` or `end` is not supported
            (not Timestamp or not supported by `np.arange`).
        """
        if isinstance(start, Timestamp) and isinstance(end, Timestamp):
            return pd.date_range(start, end, freq=self.interval)
        else:
            try:
                return np.arange(start, end, self.interval)
            except TypeError:
                raise TypeError(
                    f"Please override self.range() to support {type(start)} and {type(end)}"
                )

    @property
    def delay(self) -> float:
        """Delay between chunks in seconds. (Alias of `self.delay_seconds`.)"""
        return self.delay_seconds

    @delay.setter
    def delay(self, value: float) -> None:
        """Delay between chunks in seconds. (Alias of `self.delay_seconds`.)"""
        self.delay_seconds = value

    @abstractmethod
    async def get_one(
        self, start: TIndex, *args: PGet.args, **kwargs: PGet.kwargs
    ) -> DataFrame:
        """Get one chunk of historical data. Override this method to implement the logic.

        Parameters
        ----------
        start : TIndex
            The start index of historical data.

        Returns
        -------
        DataFrame
            The chunk of historical data.
            It is recommended to set index to Timestamp or unique incremental number.
            If the index is not Timestamp,
            override `self.to_update()` to implement the logic as well.
            If MultiIndex is used, the first level will be passed to this method
            or `self.to_update()`.
        """

    async def get(
        self, start: TIndex | None, *args: PGet.args, **kwargs: PGet.kwargs
    ) -> DataFrame:
        """Get historical data. This method does not need to be overridden.

        Parameters
        ----------
        start : TIndex
            The last index of historical data.

        Returns
        -------
        DataFrame
            The historical data.
            It is recommended to set index to Timestamp or unique incremental number.
            If the index is not Timestamp,
            override `self.to_update()` to implement the logic as well.
        """
        start_index: TIndex = start or self.start_index
        dfs = []

        # The progress bar is not accurate because chunk size may not be fixed.
        try:
            new_indices = self.new_indices(
                start_index, self.end_index, self.df_old.index
            )
        except Exception as e:
            warnings.warn(
                f"self.new_indices() failed. Progress bar may be inaccurate: {e}"
            )
            new_indices = []

        pbar = tqdm(new_indices)
        start_current: TIndex = start_index
        while self.to_update(start_current, *args, **kwargs):
            df = await self.get_one(start_current, *args, **kwargs)
            if not isinstance(df, DataFrame):
                raise TypeError(f"get_one must return DataFrame: {type(df)}")

            dfs.append(df)
            start_current = df.index.max()
            if isinstance(start_current, tuple):
                start_current = start_current[0]
            if self.add_interval_to_start_index:
                start_current += self.interval  # type: ignore
            pbar.update()
            pbar.set_description(
                f"{self.__class__.__name__} {start_current}"
                f"|{' '.join(map(str, args))}|{' '.join([f'{k}={v}' for k, v in kwargs.items()])}"
            )
            await asyncio.sleep(self.delay_seconds)
        return concat(dfs) if len(dfs) > 0 else DataFrame()


class HistoricalDataCacheWithFixedChunk(
    HistoricalDataCacheWithChunk[TIndex, TInterval, PGet],
    Generic[TIndex, TInterval, PGet],
    metaclass=ABCMeta,
):
    """Base class for historical data cache with chunk.

    This class only supports fixed interval.
    To support variable interval, use `HistoricalDataCacheWithChunk` instead.

    As `HistoricalDataCacheWithChunk` calls `self.get_one()` one by one,
    `HistoricalDataCacheWithFixedChunk` calls `self.get_one()` in parallel.
    This makes it impossible to guarantee that rate limits are not exceeded, because
    depending on network conditions etc.,
    it might theoretically be possible for all the requests to reach the server at the same time.
    Make sure to set `self.delay_seconds` large enough to avoid server overload or ban.

    Usage
    -----
    1. Override `self.get_one()` to implement the logic.
    2. Call `self.update()` to get historical data.

    Examples
    --------
    .. code-block:: python
        from cached_historical_data_fetcher import HistoricalDataCacheWithFixedChunk
        from pandas import DataFrame, Timedelta, Timestamp

        class MyCacheWithFixedChunk(HistoricalDataCacheWithFixedChunk[Timestamp, Timedelta, Any]):
            delay_seconds = 0.0
            interval = Timedelta(days=1)
            start_index = Timestamp.utcnow().floor("10D")

            async def get_one(self, start: Timestamp, *args: Any, **kwargs: Any) -> DataFrame:
                return DataFrame({"day": [start.day]}, index=[start])

        df = await MyCacheWithFixedChunk().update()
    """

    async def get(
        self, start: TIndex | None, *args: PGet.args, **kwargs: PGet.kwargs
    ) -> DataFrame:
        start_index: TIndex = start or self.start_index
        tasks = []

        pbar = tqdm(self.new_indices(start_index, self.end_index, self.df_old.index))
        for start_current in pbar:
            tasks.append(
                asyncio.create_task(self.get_one(start_current, *args, **kwargs))
            )
            pbar.update()
            pbar.set_description(
                f"{self.__class__.__name__} {start_current}"
                f"|{' '.join(map(str, args))}|{' '.join([f'{k}={v}' for k, v in kwargs.items()])}"
            )
            await asyncio.sleep(self.delay_seconds)
        dfs = await asyncio.gather(*tasks)
        return concat(dfs) if len(dfs) > 0 else DataFrame()
